﻿using System.Collections.Generic;
using System.Linq;
using System.Xml;
using System.Xml.Linq;
using SDL.Web.UI.PublicationMappings.Model.Extensions;
using SDL.Web.UI.PublicationMappings.Model.Services.DTOs;
using SDL.Web.UI.PublicationMappings.Model.Services.Implementation;
using SDL.Web.UI.PublicationMappings.Model.Services.Interfaces;
using Tridion.Web.CMUtils;
using Tridion.Web.UI.Core.Extensibility;

namespace SDL.Web.UI.PublicationMappings.DataExtenders.Extenders
{
    public class PublicationExtender
    {
        private readonly ICMSession _cmSession;
        private readonly ITopologyMappings _topologyMappings;

        private readonly string _publicationId;

        public PublicationExtender(string publicationId)
        {
            _publicationId = publicationId;

            _cmSession = CMSession.GetInstance();
            _topologyMappings = new TopologyMappings();
        }

        internal PublicationExtender(string publicationId, ICMSession cmSession, ITopologyMappings topologyMappings)
        {
            _publicationId = publicationId;

            _cmSession = cmSession;
            _topologyMappings = topologyMappings;
        }

        public XmlTextReader ProcessRequest(string command, XmlTextReader reader, PipelineContext context)
        {
            if (command == "SaveItem")
            {
                return HandlePreSaveItem(reader, context);
            }

            return reader;
        }

        public XmlTextReader ProcessResponse(string command, XmlTextReader reader, PipelineContext context)
        {
            switch (command)
            {
                case "GetItem":
                    return HandlePostGetItem(reader, context);
                case "SaveItem":
                    return HandlePostSaveItem(reader, context);
                default:
                    return reader;
            }
        }
        private XmlTextReader HandlePostGetItem(XmlTextReader reader, PipelineContext context)
        {
            return ProcessPublicationResponse(reader, context);
        }

        private XmlTextReader HandlePostSaveItem(XmlTextReader reader, PipelineContext context)
        {
            if (context.Parameters.ContainsKey("pm:mappings"))
            {
                var mappingsXmlString = (string)context.Parameters["pm:mappings"];

                var mappingsXml = new XmlDocument();
                mappingsXml.LoadXml(mappingsXmlString);

                _topologyMappings.UpdateMappings(mappingsXml);
            }

            return ProcessPublicationResponse(reader, context);
        }

        private XmlTextReader HandlePreSaveItem(XmlTextReader reader, PipelineContext context)
        {
            AssertRights(_publicationId);

            var document = reader.ToXmlDocument();

            var mappingsNode = document.SelectSingleNode("/tcm:Publication/tcm:Data/pm:Mappings", _topologyMappings.XmlNamespaceManager);

            if (mappingsNode != null && mappingsNode.ChildNodes.Count > 0)
            {
                var existingMappingsXml = _topologyMappings.GetMappingsXml(_publicationId).ToXmlElement();

				AssertDataAccessibility(existingMappingsXml);
                AssertItemsToRemove(mappingsNode, existingMappingsXml);
                AssertItemsToAdd(mappingsNode, existingMappingsXml);
                AssertDataConsistency(mappingsNode, existingMappingsXml);

				var existingMappings = _topologyMappings.GetMappings();
				AssertMappingsUrls(mappingsNode, existingMappings);

                context.Parameters.Add("pm:mappings", mappingsNode.OuterXml);
                mappingsNode.ParentNode.RemoveChild(mappingsNode);
            }

            return document.ToReader();
        }

        private void AssertRights(string publicationId)
        {
	        if (!_cmSession.AreMappingsEditable(publicationId))
	        {
		        throw new NoRightsException(_cmSession.UserId, publicationId);
	        }
        }

	    private void AssertDataAccessibility(XmlNode mappingsNode)
        {
            var mappingsAccessibilityAttribute = mappingsNode.Attributes["areMappingsAccessible"];
            if (mappingsAccessibilityAttribute != null && mappingsAccessibilityAttribute.Value == "false")
            {
                throw new TopologyManagerUnavailableException();
            }
        }

		private static void AssertItemsToAdd(XmlNode newMappingsNode, XmlNode existingMappingsNode)
		{
			var nodes = newMappingsNode.ChildNodes.Cast<XmlNode>().ToArray();

			var environmentsToAdd = nodes
				.Where(x => x.Attributes["Id"] == null)
				.Select(x => x.Attributes["EnvironmentPurpose"].Value);
			
			var environmentsToRemove = nodes
				.Where(x => x.Attributes["ToRemove"] != null && x.Attributes["ToRemove"].Value == "true")
				.Select(x => x.Attributes["EnvironmentPurpose"].Value);

			var existingEnvironments = GetMappingsWithId(existingMappingsNode)
				.Select(x => x.Attributes["EnvironmentPurpose"].Value);

			var environmentIsConfigured = environmentsToAdd
				.Where(purpose => !environmentsToRemove.Contains(purpose))
				.Any(purpose => existingEnvironments.Contains(purpose));

			if (environmentIsConfigured)
			{
				throw new EnvironmentIsAlreadyConfiguredException();
			}
		}
		private static void AssertItemsToRemove(XmlNode newMappingsNode, XmlNode existingMappingsNode)
		{
			var mappingsToRemoveIds = GetMappingsWithId(newMappingsNode)
				.Where(x => x.Attributes["ToRemove"] != null && x.Attributes["ToRemove"].Value == "true")
				.Select(x => x.Attributes["Id"].Value);

			var existingMappingsWithPublishedItems = GetMappingsWithId(existingMappingsNode)
				.Where(x=> x.Attributes["HasPublishedItems"] !=null && x.Attributes["HasPublishedItems"].Value == "true")
				.Select(x => x.Attributes["Id"].Value);

			var mappingsToRemoveAreInUse = mappingsToRemoveIds.Any(id => existingMappingsWithPublishedItems.Contains(id));

			if (mappingsToRemoveAreInUse)
			{
				throw new UnableRemoveMappingsException();
			}
		}

		private void AssertDataConsistency(XmlNode mappingsNode, XmlElement existingMappings)
        {
            var existingMappingsIds = GetMappingsWithId(existingMappings).Select(x => x.Attributes["Id"].Value);
			var newMappingsIds = GetMappingsWithId(mappingsNode)
				.Where(mapping => mapping.Attributes["ToRemove"] == null || mapping.Attributes["ToRemove"].Value == "false")
				.Select(x => x.Attributes["Id"].Value);

	        var hasInconsistentMappings = newMappingsIds.Any(id => !existingMappingsIds.Contains(id));

            if (hasInconsistentMappings)
            {
               throw new EditDeleteMappingException();
            }
        }

	    private static IEnumerable<XmlNode> GetMappingsWithId(XmlNode mappings)
	    {
		    return mappings.ChildNodes.Cast<XmlNode>().Where(x => x.Attributes["Id"] != null);
	    }

	    private static void AssertMappingsUrls(XmlNode mappingsNode, IEnumerable<TopologyMapping> existingMappings)
        {
			var mappingsIds = GetMappingsWithId(mappingsNode)
				.Where(x=> x.Attributes["ToRemove"] == null)
				.Select(x => x.Attributes["Id"].Value);

		    var mappingsWithUrl = mappingsNode
			    .ChildNodes.Cast<XmlNode>()
			    .Where(x => x.Attributes["BaseMappedUrl"] != null && x.Attributes["RelativeUrl"] != null);

		    var mappingsUrls = mappingsWithUrl
				.Where(x => x.Attributes["ToRemove"] == null)
				.Select(mappingNode => mappingNode.Attributes["BaseMappedUrl"].Value + "/" + mappingNode.Attributes["RelativeUrl"].Value);
			
			var removedMappingsUrls = mappingsWithUrl
				.Where(x => x.Attributes["ToRemove"] != null && x.Attributes["ToRemove"].Value == "true")
				.Select(mappingNode => mappingNode.Attributes["BaseMappedUrl"].Value + "/" + mappingNode.Attributes["RelativeUrl"].Value);

	        var existingMappingsUrls = existingMappings
				.Where(mapping => !mappingsIds.Contains(mapping.Id))
				.Select(mapping => mapping.PrimaryMappedUrl)
				.Except(removedMappingsUrls);

			var hasInconsistentUrls = mappingsUrls.Any(url => existingMappingsUrls.Contains(url));

            if (hasInconsistentUrls)
            {
                throw new MappingsUrlsInUseException();
            }
        }

        private XmlTextReader ProcessPublicationResponse(XmlTextReader reader, PipelineContext context)
        {
            var document = reader.ToXmlDocument();

            var mappingsXml = _topologyMappings.GetMappingsXml(_publicationId);
            var areMappingsEditable = _cmSession.AreMappingsEditable(_publicationId);
            AddMappingsData(document, mappingsXml, areMappingsEditable);

            var bptReadOnly = !_cmSession.HasBptManagmentRights(_publicationId);
            SetBptReadOnlyOption(document, bptReadOnly);

            return document.ToReader();
        }

        private void SetBptReadOnlyOption(XmlDocument document, bool bptReadOnly)
        {
            var bptNode = document.SelectSingleNode("/tcm:Publication/tcm:Data/tcm:BusinessProcessType", _topologyMappings.XmlNamespaceManager);
            if (bptNode == null)
            {
                return;
            }

            var attribute = document.CreateAttribute("ReadOnly");
            attribute.Value = bptReadOnly.ToString().ToLower();

            bptNode.Attributes.Append(attribute);
        }

        private void AddMappingsData(XmlDocument document, XElement mappingsXml, bool areMappingsEditable)
        {
            var dataNode = document.SelectSingleNode("/tcm:Publication/tcm:Data", _topologyMappings.XmlNamespaceManager);

            if (dataNode == null)
            {
                return;
            }

            mappingsXml.SetAttributeValue("IsEditable", areMappingsEditable);

            var node = mappingsXml.ToXmlElement();

            var importNode = dataNode.OwnerDocument.ImportNode(node, true);
            dataNode.AppendChild(importNode);
        }
    }
}