﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text.RegularExpressions;
using System.Xml;
using System.Xml.Linq;
using Microsoft.OData.Client;
using SDL.Web.UI.PublicationMappings.Model.Services.DTOs;
using Tridion.TopologyManager.Client;
using Tridion.Web.CMUtils;
using SDL.Web.UI.PublicationMappings.Model.Services.Interfaces;
using Tridion.ContentManager.CoreService.Client;

namespace SDL.Web.UI.PublicationMappings.Model.Services.Implementation
{
	public class TopologyMappings : ITopologyMappingsService, ITopologyMappings
	{
		private readonly ITopologyManagerClient _tmClient;
		private readonly ISessionAwareCoreService _coreService;

		internal TopologyMappings(ITopologyManagerClient tmClient, ISessionAwareCoreService coreService)
		{
			_tmClient = tmClient;
			_coreService = coreService;
		}

		public TopologyMappings()
		{
			_tmClient = new TopologyManagerClientFactory().Create();
			_coreService = CMSession.GetInstance().CoreServiceClient;
		}

		private XmlNamespaceManager _xmlNamespaceManager;
		public XmlNamespaceManager XmlNamespaceManager
		{
			get
			{
				if (_xmlNamespaceManager == null)
				{
					var xmlNamespaceManager = new XmlNamespaceManager(new NameTable());

					xmlNamespaceManager.AddNamespace(Tridion.Constants.TcmPrefix, Tridion.Constants.TcmNamespace);
					xmlNamespaceManager.AddNamespace(Tridion.Constants.XlinkPrefix, Tridion.Constants.XlinkNamespace);
					xmlNamespaceManager.AddNamespace(Constants.PublicationMappingsPrefix, Constants.PublicationMappingsNamespace.NamespaceName);

					_xmlNamespaceManager = xmlNamespaceManager;
				}

				return _xmlNamespaceManager;
			}
		}

		public string GetMappings(string publicationId)
		{
			return GetMappingsXml(publicationId).ToString();
		}

		public XElement GetMappingsXml(string publicationId)
		{
			var mappings = new MappingData[0];

			try
			{
				mappings = _tmClient.Mappings
					.Expand("CdEnvironment")
					.Where(mapping =>
						mapping.CmEnvironmentId == TopologyManagerClient.ContentManagerEnvironmentId
						&& mapping.PublicationId == publicationId)
					.ToArray();
			}
			catch (InvalidOperationException)
			{
				return ConvertMappingsToXml(mappings, false);
			}

			return ConvertMappingsToXml(mappings);
		}

		public IEnumerable<TopologyMapping> GetMappings()
		{
			try
			{
				return _tmClient.Mappings
					.Where(mapping => mapping.CmEnvironmentId == TopologyManagerClient.ContentManagerEnvironmentId)
					.Select(data => new TopologyMapping
					{
						Id = data.Id,
						WebApplicationId = data.WebApplicationId,
						CmEnvironmentId = data.CmEnvironmentId,
						EnvironmentPurpose = data.EnvironmentPurpose,
						PrimaryMappedUrl = data.PrimaryMappedUrl,
						RelativeUrl = data.RelativeUrl,
						PublicationId = data.PublicationId,
						IsOffline = data.IsOffline
					});
			}
			catch (InvalidOperationException ex)
			{
				throw TopologyManagerClient.DataServiceException.CreateParsedException(ex);
			}
		}

		private XElement ConvertMappingsToXml(IEnumerable<MappingData> mappings, bool areMappingsAccessible = true)
		{
			var listElement = ConvertListToXml("Mappings", mappings.Select(ConvertMappingToXml));
			listElement.SetAttributeValue("areMappingsAccessible", areMappingsAccessible);
			return listElement;
		}

		private XElement ConvertMappingToXml(MappingData mapping)
		{
			return new XElement(Constants.PublicationMappingsNamespace + "Mapping",
				new XAttribute("Id", mapping.Id),
				new XAttribute("WebApplicationId", mapping.WebApplicationId),
				new XAttribute("EnvironmentPurpose", mapping.EnvironmentPurpose),
				new XAttribute("BaseMappedUrl", Regex.Replace(mapping.PrimaryMappedUrl, mapping.RelativeUrl + "$", string.Empty)),
				new XAttribute("RelativeUrl", mapping.RelativeUrl),
				new XAttribute("PublicationId", mapping.PublicationId),
				new XAttribute("HasPublishedItems", HasPublishedItems(mapping)),
				mapping.CdEnvironment == null
					? null
					: new XAttribute("CdEnvironmentId", mapping.CdEnvironment.Id)
				);
		}

		private bool HasPublishedItems(MappingData mapping)
		{
			var filterData = new PublishedItemsFilterData
			{
				BaseColumns = ListBaseColumns.Id,
				ForPublication = new LinkToPublicationData { IdRef = mapping.PublicationId },
				// TODO AN: temporaly check if publication has any published items due to CRQ-1553
				// Purpose = mapping.EnvironmentPurpose, 
				ResultLimit = 1
			};

			return _coreService.GetSystemWideList(filterData).Any();
		}

		public void UpdateMappings(XmlDocument mappingsXml)
		{
			if (mappingsXml == null)
			{
				return;
			}

			try
			{
				var mappingNodes = mappingsXml.SelectNodes("/pm:Mappings/pm:Mapping", XmlNamespaceManager);

				if (mappingNodes.Count != 0)
				{
					foreach (XmlNode mappingNode in mappingNodes)
					{
						UpdateMapping(mappingNode);
					}

					_tmClient.SaveChanges(SaveChangesOptions.BatchWithSingleChangeset);
				}
			}

			catch (InvalidOperationException ex)
			{
				throw TopologyManagerClient.DataServiceException.CreateParsedException(ex);
			}
		}

		private void UpdateMapping(XmlNode mappingNode)
		{
			var id = mappingNode.Attributes["Id"] != null ? mappingNode.Attributes["Id"].Value : String.Empty;
			if (String.IsNullOrEmpty(id))
			{
				var newMappingData = new MappingData
				{
					Id = id,
					RelativeUrl = mappingNode.Attributes["RelativeUrl"].Value,
					WebApplicationId = mappingNode.Attributes["WebApplicationId"].Value,
					PublicationId = mappingNode.Attributes["PublicationId"].Value,
					CmEnvironmentId = TopologyManagerClient.ContentManagerEnvironmentId,
					ExtensionProperties = new KeyedByTypeCollection<ExtensionProperty>()
				};

				_tmClient.AddToMappings(newMappingData);
				return;
			}

			var mapping = _tmClient.Mappings.SingleOrDefault(x => x.Id == id);
			if (mapping == null)
			{
				return;
			}

			if (mappingNode.Attributes["ToRemove"] != null)
			{
				_tmClient.DeleteObject(mapping);
				return;
			}

			mapping.RelativeUrl = mappingNode.Attributes["RelativeUrl"].Value;
			mapping.WebApplicationId = mappingNode.Attributes["WebApplicationId"].Value;

			_tmClient.UpdateObject(mapping);
		}

		public string GetTopologies(string topologyTypeId, string[] webApplicationIds = null)
		{
			try
			{
				var topologies = FindTopologies(topologyTypeId, webApplicationIds);
				var items = topologies.Select(ConvertTopologyToXml);
				return ConvertListToXml("CdTopologies", items).ToString();
			}
			catch (InvalidOperationException ex)
			{
				throw TopologyManagerClient.DataServiceException.CreateParsedException(ex);
			}
		}

		private IEnumerable<CdTopologyData> FindTopologies(string topologyTypeId, string[] webApplicationIds = null)
		{
			var topologies = _tmClient.CdTopologies.Where(x => x.CdTopologyTypeId == topologyTypeId);
			if (webApplicationIds == null)
			{
				return topologies;
			}

			var cdEnvironmentIds = _tmClient.WebApplications.Expand("CdEnvironment")
				.Where(x => webApplicationIds.Contains(x.Id))
				.Select(x => x.CdEnvironment.Id);

			return topologies.Where(x => x.CdEnvironmentIds.Intersect(cdEnvironmentIds).Any());
		}

		private static XElement ConvertTopologyToXml(CdTopologyData topology)
		{
			var result = new XElement(Constants.PublicationMappingsNamespace + "CdTopology",
				new XAttribute("Id", topology.Id),
				new XAttribute("Name", topology.Name));

			foreach (var cdEnvironmentId in topology.CdEnvironmentIds)
			{
				var cdEnvironment = new XElement(Constants.PublicationMappingsNamespace + "CdEnvironment", new XAttribute("Id", cdEnvironmentId));
				result.Add(cdEnvironment);
			}

			return result;
		}

		public string GetWebApplications(string[] environmentIds, string purpose = null)
		{
			try
			{
				var webAppUrlById = new Dictionary<string, string>();

				var webApps = _tmClient.WebApplications.Expand("CdEnvironment")
					.Where(x => environmentIds.Contains(x.CdEnvironment.Id) && (purpose == null || purpose == x.EnvironmentPurpose));

				var webSites = _tmClient.Websites;

				foreach (var webApp in webApps)
				{
					var webSite = webSites.Single(x => x.Id == webApp.WebsiteId);
					webAppUrlById.Add(webApp.Id, webSite.BaseUrls[0] + webApp.ContextUrl);
				}

				var items = webAppUrlById.Select(x => ConvertWebApplicationToXml(x.Key, x.Value));
				return ConvertListToXml("WebApplications", items).ToString();
			}
			catch (InvalidOperationException ex)
			{
				throw TopologyManagerClient.DataServiceException.CreateParsedException(ex);
			}
		}

		private static XElement ConvertWebApplicationToXml(string id, string url)
		{
			return new XElement(Constants.PublicationMappingsNamespace + "WebApplication", new XAttribute("Id", id))
			{
				Value = url
			};
		}

		private static XElement ConvertListToXml(string nodeName, IEnumerable<XElement> items)
		{
			var listElement = new XElement(Constants.PublicationMappingsNamespace + nodeName,
				new XAttribute(XNamespace.Xmlns + Constants.PublicationMappingsPrefix, Constants.PublicationMappingsNamespace.NamespaceName));

			listElement.Add(items);

			return listElement;
		}
	}
}
