﻿using System.Collections.Generic;
using NSubstitute;
using NUnit.Framework;
using SDL.Web.UI.PublicationMappings.DataExtenders;
using SDL.Web.UI.PublicationMappings.DataExtenders.Extenders;
using SDL.Web.UI.PublicationMappings.Model.Services.Interfaces;
using Tridion.ContentManager.CoreService.Client;
using Tridion.Web.CMUtils;

namespace SDl.Web.UI.PublicationMappings.Tests.DataExtenders.BusinessProcessTypeList
{
    [TestFixture]
    public class PostGetListFixture
    {
        private BusinessProcessTypeListExtender _extender;
        private ITopologyMappings _topologyMappings;
        private ICMSession _cmSession;
        private ISessionAwareCoreService _coreService;

        [SetUp]
        public void SetUp()
        {
            _cmSession = Substitute.For<ICMSession>();
            _coreService = Substitute.For<ISessionAwareCoreService>();

            _cmSession.CoreServiceClient.Returns(_coreService);

            _topologyMappings = Substitute.For<ITopologyMappings>();

            _topologyMappings.XmlNamespaceManager.Returns(DataExtenderTestsHelper.XmlNamespaceManager);

            _extender = new BusinessProcessTypeListExtender(_cmSession, _topologyMappings);
        }

        private static string GetListXmlString(IEnumerable<string> bptIds)
        {
            var result = "<tcm:ListBusinessProcessTypes xmlns:tcm=\"http://www.tridion.com/ContentManager/5.0\">";
            foreach (var id in bptIds)
            {
                result += "<tcm:Item ID=\"" + id + "\" Title=\"BPT\" Icon=\"T4096L0P0\" Type=\"4096\" IsNew=\"false\" Modified=\"2015-12-02T15:40:23\" OwningPublicationID=\"6\" OwningPublicationTitle=\"400 Master WebSite\" />";
            }
            result += "</tcm:ListBusinessProcessTypes>";
            return result;
        }

        [Test]
        public void Should_not_chang_xml_if_list_is_not_BusinessProcessTypes()
        {
            var input = "<tcm:FakeList xmlns:tcm=\"http://www.tridion.com/ContentManager/5.0\">" +
                            "<tcm:Item ID=\"tcm:6-4-4096\" Title=\"BPT\" Icon=\"T4096L0P0\" Type=\"4096\" IsNew=\"false\" Modified=\"2015-12-02T15:40:23\" OwningPublicationID=\"6\" OwningPublicationTitle=\"400 Master WebSite\" />" +
                        "</tcm:FakeList>";

            var reader = DataExtenderTestsHelper.GetReader(input);
            var context = DataExtenderTestsHelper.GetPipelineContext("GetList");

            var result = _extender.HandlePostGetList(reader, context);

            Assert.AreEqual(input, result.ToXmlDocument().InnerXml);
        }

        [Test]
        public void Should_add_purposes_to_list()
        {
            var bpt1 = new BusinessProcessTypeData
            {
                Id = "tcm:6-4-4096",
                TargetTypes = new[]
                {
                    new TargetTypeData {Purpose = "Staging"},
                    new TargetTypeData {Purpose = "Live"}
                },
				CdTopologyTypeId = "TopologyType1"
            };

            var bpt2 = new BusinessProcessTypeData
            {
                Id = "tcm:6-5-4096",
                TargetTypes = new[]
                {
                    new TargetTypeData { Purpose = "Live"},
                    new TargetTypeData { Purpose = "Acceptance"}
                }
            };

            var input = GetListXmlString(new[] { bpt1.Id, bpt2.Id });

            var reader = DataExtenderTestsHelper.GetReader(input);
            var context = DataExtenderTestsHelper.GetPipelineContext("GetList");

            _cmSession.CoreServiceClient.Read(bpt1.Id, Arg.Any<ReadOptions>()).Returns(bpt1);
            _cmSession.CoreServiceClient.Read(bpt2.Id, Arg.Any<ReadOptions>()).Returns(bpt2);

            var result = _extender.HandlePostGetList(reader, context);

            var expected = "<tcm:ListBusinessProcessTypes xmlns:tcm=\"http://www.tridion.com/ContentManager/5.0\">" +
								"<tcm:Item ID=\"tcm:6-4-4096\" Title=\"BPT\" Icon=\"T4096L0P0\" Type=\"4096\" IsNew=\"false\" Modified=\"2015-12-02T15:40:23\" OwningPublicationID=\"6\" OwningPublicationTitle=\"400 Master WebSite\" CdTopologyTypeId=\"TopologyType1\">" +
									"<pm:Purpose xmlns:pm=\"http://www.sdltridion.com/2015/PublicationMappings\">Staging</pm:Purpose>" +
									"<pm:Purpose xmlns:pm=\"http://www.sdltridion.com/2015/PublicationMappings\">Live</pm:Purpose>" +
                                "</tcm:Item>" +
								"<tcm:Item ID=\"tcm:6-5-4096\" Title=\"BPT\" Icon=\"T4096L0P0\" Type=\"4096\" IsNew=\"false\" Modified=\"2015-12-02T15:40:23\" OwningPublicationID=\"6\" OwningPublicationTitle=\"400 Master WebSite\" CdTopologyTypeId=\"\">" +
									"<pm:Purpose xmlns:pm=\"http://www.sdltridion.com/2015/PublicationMappings\">Live</pm:Purpose>" +
									"<pm:Purpose xmlns:pm=\"http://www.sdltridion.com/2015/PublicationMappings\">Acceptance</pm:Purpose>" +
                                "</tcm:Item>" +
                            "</tcm:ListBusinessProcessTypes>";

            Assert.AreEqual(expected, result.ToXmlDocument().InnerXml);
        }
    }
}
