﻿using System;
using System.Linq;
using System.Text.RegularExpressions;
using System.Xml.Linq;
using NSubstitute;
using NUnit.Framework;
using SDL.Web.UI.PublicationMappings.DataExtenders;
using SDL.Web.UI.PublicationMappings.Model.Services.DTOs;
using Tridion.ContentManager;
using Tridion.Web.UI.Core.Extensibility;

namespace SDl.Web.UI.PublicationMappings.Tests.DataExtenders.Publication
{
    [TestFixture]
    public class PreSavePublicationFixture : PublicationExtenderFixtureBase
    {
        private const string CommandName = "SaveItem";
        private PipelineContext _context;

		[SetUp]
        public new void SetUp()
        {
            base.SetUp();

            _context = DataExtenderTestsHelper.GetPipelineContext(CommandName, PublicationId);
            
            MockBptId(PublicationId, TcmUri.UriNull);
        }

		private static string GetInputPublicationXmlString(TestTopologyData[] mappings)
		{
			var result =
				"<tcm:Publication ID=\"tcm:0-5-1\" xmlns:tcm=\"http://www.tridion.com/ContentManager/5.0\" xmlns:xlink=\"http://www.w3.org/1999/xlink\" xmlns:i=\"http://www.w3.org/2001/XMLSchema-instance\">" +
					"<tcm:Info />" +
					"<tcm:Data>" +
					GetMappingsString(mappings) +
					"</tcm:Data>" +
					"<tcm:Rights />" +
				"</tcm:Publication>";

			return result;
		}


	    private static string GetMappingsString(TestTopologyData[] mappings)
	    {
			var result = (mappings.Any() ? "<pm:Mappings xmlns:pm=\"http://www.sdltridion.com/2015/PublicationMappings\">" : string.Empty);

			foreach (var mapping in mappings)
			{
				result += GetMappingString(mapping);
			}

		    result += (mappings.Any() ? "</pm:Mappings>" : string.Empty);

		    return result;
	    }

		private static string GetMappingString(TestTopologyData mapping)
		{
			return String.Format(
				"<pm:Mapping" + (string.IsNullOrEmpty(mapping.Id) ? "" : " Id=\"{0}\"") + 
					" WebApplicationId=\"{1}\" EnvironmentPurpose=\"{2}\" BaseMappedUrl=\"{3}\" RelativeUrl=\"{4}\""
					+ (mapping.ToRemove ? " ToRemove=\"true\"" : "") 
					+ " HasPublishedItems=\"{5}\" />", 

				mapping.Id, 
				mapping.WebApplicationId, 
				mapping.EnvironmentPurpose, 
				mapping.BaseMappedUrl, 
				mapping.RelativeUrl,
				mapping.HasPublishedItems.ToString().ToLower());
		}


        private static TestTopologyData[] GetInputMappings()
        {
			return new[]
			{
				MappingData("Mapping_id_1", "Live web app", "Live", "http://test.com/test", "/test"),
				MappingData("Mapping_id_2", "Live web app", "Live", "http://testnew.com/new", "/new"),
				MappingData("Other_Mapping_id", "Live web app", "Live", "http://testold.com/old", "/old")
			};
        }


	    [Test]
        public void Should_not_change_xml_and_context_if_there_is_no_custom_data()
        {
            var input = GetInputPublicationXmlString(new TestTopologyData[0]);
            var reader = DataExtenderTestsHelper.GetReader(input);

            var result = Extender.ProcessRequest(CommandName, reader, _context);

            Assert.AreEqual(input, result.ToXmlDocument().InnerXml);
            Assert.IsFalse(_context.Parameters.ContainsKey("pm:mappings"));
        }

        [Test]
        public void Should_remove_custom_data_from_input_xml()
        {
	        var mappings = GetInputMappings().ToArray();

	        var input = GetInputPublicationXmlString(mappings);
            var reader = DataExtenderTestsHelper.GetReader(input);

            var expected = GetInputPublicationXmlString(new TestTopologyData[0]);

            var existingMappingsXml = XElement.Parse(GetMappingsString(mappings));
            TopologyMappings.GetMappingsXml(PublicationId).Returns(existingMappingsXml);

            var result = Extender.ProcessRequest(CommandName, reader, _context);

            Assert.AreEqual(expected, result.ToXmlDocument().InnerXml);
        }

        [Test]
        public void Should_put_custom_data_to_context()
        {
			var mappings = GetInputMappings().ToArray();

			var input = GetInputPublicationXmlString(mappings);
			var reader = DataExtenderTestsHelper.GetReader(input);

            var expected = GetMappingsString(mappings);

			var existingMappingsXml = XElement.Parse(expected);
			TopologyMappings.GetMappingsXml(PublicationId).Returns(existingMappingsXml);

            Extender.ProcessRequest(CommandName, reader, _context);

            Assert.AreEqual(expected, _context.Parameters["pm:mappings"]);
        }

        [Test]
        public void Should_throw_exception_if_topology_manager_unavailable()
        {
	        var input = GetInputPublicationXmlString(GetInputMappings());
			var reader = DataExtenderTestsHelper.GetReader(input);

            var existingMappingsXml = XElement.Parse("<pm:Mappings xmlns:pm=\"http://www.sdltridion.com/2015/PublicationMappings\" areMappingsAccessible=\"false\"/>");
            TopologyMappings.GetMappingsXml(PublicationId).Returns(existingMappingsXml);

            Assert.Throws<TopologyManagerUnavailableException>(() => Extender.ProcessRequest(CommandName, reader, _context));
        }

        [Test]
        public void Should_throw_exception_if_user_has_no_rights()
        {
	        var input = GetInputPublicationXmlString(GetInputMappings());
			var reader = DataExtenderTestsHelper.GetReader(input);

            CmSession.UserId.Returns("TestUser");

            MockRights(false, false);

            Assert.Throws<NoRightsException>(() => Extender.ProcessRequest(CommandName, reader, _context));
        }
		
		[Test, TestCaseSource("ThrowExceptionCases")]
		public void Should_throw_exception(string caseName, TestTopologyData[] newMappings, TestTopologyData[] existingMappings, Type exceptionType)
		{
			var input = GetInputPublicationXmlString(newMappings);

			var reader = DataExtenderTestsHelper.GetReader(input);

			var existingMappingsXml = XElement.Parse(GetMappingsString(existingMappings));
			TopologyMappings.GetMappingsXml(PublicationId).Returns(existingMappingsXml);

			TopologyMappings.GetMappings().Returns(existingMappings.Select(x=>x.ToTopologyMapping()));

			Assert.Throws(exceptionType, () => Extender.ProcessRequest(CommandName, reader, _context));
		}

	    public static object[] ThrowExceptionCases =
	    {
			new object[]
		    {
				"When new mapping urls conflict with existing",
			    new[]
			    {
				    MappingData("", "WebApplicationId3", "Staging", "http://test.com/test", "test", baseMappedUrl: "http://test.com"),
				    MappingData("", "WebApplicationId4", "Live", "http://testnew.com/new", "new", baseMappedUrl: "http://testnew.com")
			    },
			    new[]
			    {
				    MappingData("Mapping_id_1", "WebApplicationId1", "Cloud", "http://test.com/test", "/test"),
				    MappingData("Mapping_id_2", "WebApplicationId2", "Other", "http://testnew.com/new", "/new")
			    },
				typeof(MappingsUrlsInUseException)
			},
			new object[]
		    {
				"When remove mapping with published items",
			    new[]
			    {
					MappingData("Mapping_id_1", "WebApplicationId1", "Staging", "http://test.com/test", "/test", toRemove: true)
				},
			    new[]
			    {
					MappingData("Mapping_id_1", "WebApplicationId1", "Staging", "http://test.com/test", "/test", hasPublishedItems: true)
				},
				typeof(UnableRemoveMappingsException)
			},
			new object[]
		    {
				"When add mapping for existing purpose",
			    new[]
			    {
					MappingData("", "WebApplicationId1", "Staging", "http://test.com/test", "/test"),
				},
			    new[]
			    {
					MappingData("Mapping_id_1", "WebApplicationId1", "Staging", "http://test.com/test", "/test"),
				},
				typeof(EnvironmentIsAlreadyConfiguredException)
			},
			new object[]
		    {
				"When edit deleted mappings",
			    new[]
			    {
					MappingData("Mapping_id_1", "WebApplicationId1", "Live", "http://test.com/test", "/test"),
					MappingData("DeletedMappingId", "WebApplicationId2", "Staging", "http://testnew.com/new", "/new"),
					MappingData("Other_Mapping_id", "WebApplicationId3", "Cloud", "http://testold.com/old", "/old")
				},
			    new[]
			    {
					MappingData("Mapping_id_1", "WebApplicationId1", "Staging", "http://test.com/test", "/test"),
					MappingData("Other_Mapping_id", "SomeWebApplicationId", "Cloud", "http://testnew.com/new", "/old")
				},
				typeof(EditDeleteMappingException)
			},
			new object[]
			{
				"When edited mapping urls conflict with existing",
				new[]
				{
					MappingData("Mapping_id_1", "WebApplicationId1", "Staging", "http://conflicting.url.com/test", "test", baseMappedUrl: "http://conflicting.url.com"),
					MappingData("Mapping_id_2", "WebApplicationId2", "Live", "http://testnew.com/new", "new", baseMappedUrl: "http://testnew.com"),
				},
				new[]
				{
					MappingData("Mapping_id_1", "WebApplicationId1", "Staging", "http://first_test.com/test", "/test"),
					MappingData("Mapping_id_2", "WebApplicationId2", "Live", "http://testnew.com/new", "/new"),
					MappingData("Mapping_id_3", "WebApplicationId3", "Draft", "http://conflicting.url.com/test", "/test")
				},
				typeof(MappingsUrlsInUseException)
			},
			new object[]
		    {
				"Mixed",
				new[]
			    {
				    MappingData("Mapping_id_1", "WebApplicationId1", "Staging", "http://test.com/test", "test", baseMappedUrl: "http://test.com"),
				    MappingData("Mapping_id_2", "WebApplicationId2", "Live", "http://testnew.com/new", "new", baseMappedUrl: "http://testnew.com"),
				    MappingData("Mapping_id_3", "WebApplicationId3", "Draft", "http://testold.com/old", "old", baseMappedUrl: "http://testold.com", toRemove: true),
				    MappingData("", "SomeWebApplicationId", "Cloud", "http://testadd.com/add", "add", baseMappedUrl: "http://testadd.com")
			    },
			    new[]
			    {
				    MappingData("Mapping_id_1", "WebApplicationId1", "Staging", "http://test.com/test", "/test"),
				    MappingData("Mapping_id_2", "WebApplicationId2", "Live", "http://testnew.com/new", "/new"),
				    MappingData("Mapping_id_3", "WebApplicationId3", "Draft", "http://testold.com/old", "/old"),
				    MappingData("Mapping_id_4", "WebApplicationId4", "Live", "http://testadd.com/add", "/add")
			    },
				typeof(MappingsUrlsInUseException)
			}
	    };

		[Test, TestCaseSource("DoNotThrowExceptionCases")]
		public void Should_not_throw_exception(string caseName, TestTopologyData[] newMappings, TestTopologyData[] existingMappings)
		{
			var input = GetInputPublicationXmlString(newMappings);
			var reader = DataExtenderTestsHelper.GetReader(input);

			var existingMappingsXml = XElement.Parse(GetMappingsString(existingMappings));
			TopologyMappings.GetMappingsXml(PublicationId).Returns(existingMappingsXml);
			TopologyMappings.GetMappings().Returns(existingMappings.Select(x => x.ToTopologyMapping()));

			Assert.DoesNotThrow(() => Extender.ProcessRequest(CommandName, reader, _context));
		}

	    public static object[] DoNotThrowExceptionCases =
	    {
		    new object[]
		    {
			    "When new mappings are the same as existing",
			    new[]
			    {
				    MappingData("Mapping_id_1", "WebApplicationId1", "Staging", "http://test.com/test", "test",
					    baseMappedUrl: "http://test.com"),
				    MappingData("Mapping_id_2", "WebApplicationId2", "Live", "http://testnew.com/new", "new",
					    baseMappedUrl: "http://testnew.com"),
				    MappingData("Other_Mapping_id", "WebApplicationId3", "Draft", "http://testold.com/old", "old",
					    baseMappedUrl: "http://testold.com")
			    },
			    new[]
			    {
				    MappingData("Mapping_id_1", "WebApplicationId1", "Staging", "http://test.com/test", "/test"),
				    MappingData("Mapping_id_2", "WebApplicationId2", "Live", "http://testnew.com/new", "/new"),
				    MappingData("Other_Mapping_id", "WebApplicationId3", "Draft", "http://testold.com/old", "/old")
			    }
		    },
			new object[]
		    {
			    "When remove already removed mapping",
			    new[]
			    {
					MappingData("Mapping_id_1", "WebApplicationId1", "Staging", "http://test.com/test", "test", baseMappedUrl: "http://test.com", toRemove: true)
				},
			    new[]
			    {
					MappingData("Mapping_id_2", "WebApplicationId2", "Live", "http://test.com/test", "/test")
				}
		    },
			new object[]
		    {
			    "When remove and add the same mapping",
			    new[]
			    {
					MappingData("Mapping_id_1", "WebApplicationId1", "Staging", "http://test.com/test", "test", baseMappedUrl: "http://test.com", toRemove: true),
					MappingData("", "WebApplicationId1", "Staging", "http://test.com/test", "test", baseMappedUrl: "http://test.com")
				},
			    new[]
			    {
					MappingData("Mapping_id_1", "WebApplicationId1", "Staging", "http://test.com/test", "/test")
				}
		    }
	    };

		private static TestTopologyData MappingData(string id, string webApplicationId, string environmentPurpose, 
			string primaryMappedUrl, string relativeUrl, string cdEnvironmentId = "CdEnvironmentId", 
			string baseMappedUrl = "", bool toRemove = false, bool hasPublishedItems = false)
		{
			return new TestTopologyData
			{
				Id = id,
				WebApplicationId = webApplicationId,
				CmEnvironmentId = cdEnvironmentId,
				EnvironmentPurpose = environmentPurpose,
				PrimaryMappedUrl = primaryMappedUrl,
				RelativeUrl = relativeUrl,
				PublicationId = PublicationId,
				ToRemove = toRemove,
				HasPublishedItems = hasPublishedItems,
				BaseMappedUrl = string.IsNullOrEmpty(baseMappedUrl) 
					? Regex.Replace(primaryMappedUrl, relativeUrl + "$", string.Empty)
					: baseMappedUrl
			};
		}

	    public class TestTopologyData
	    {
			public string Id;
			public string EnvironmentPurpose;
			public string WebApplicationId;
			public string CmEnvironmentId;
			public string PrimaryMappedUrl;
			public string RelativeUrl;
			public string BaseMappedUrl;
			public bool ToRemove;
			public bool HasPublishedItems;
			public string PublicationId;

		    public TopologyMapping ToTopologyMapping()
		    {
			    return new TopologyMapping
			    {
					Id = Id,
					WebApplicationId = WebApplicationId,
					CmEnvironmentId = CmEnvironmentId,
					EnvironmentPurpose = EnvironmentPurpose,
					PrimaryMappedUrl = PrimaryMappedUrl,
					RelativeUrl = RelativeUrl,
					PublicationId = PublicationId
				};
		    }
		}
	}
}
