﻿using System.Linq;
using NSubstitute;
using NUnit.Framework;
using SDL.Web.UI.PublicationMappings.Model.Extensions;
using Tridion.ContentManager.CoreService.Client;
using Tridion.Web.CMUtils;

namespace SDl.Web.UI.PublicationMappings.Tests.Extensions
{
    [TestFixture]
    public class CmSessionExtensionsFixture
    {
        private string _publicationId = "tcm:0-6-1";
        
        private ICMSession _cmSession;
        private ISessionAwareCoreService _coreServiceClient;

        [SetUp]
        public void SetUp()
        {
            _cmSession = Substitute.For<ICMSession>();
            _coreServiceClient = Substitute.For<ISessionAwareCoreService>();

            _cmSession.CoreServiceClient.Returns(_coreServiceClient);
        }

        [Test]
        [Description("Should return true if user has bpt managment rights")]
        [TestCase(true, true, true, Description = "Administrator, Publication manager and Bpt Manager rights")]
        [TestCase(true, false, true, Description = "Administrator, NO Publication manager and Bpt Manager rights")]
        [TestCase(true, true, false, Description = "Administrator, Publication manager and NO Bpt Manager rights")]
        [TestCase(true, false, false, Description = "Administrator, NO Publication manager and NO Bpt Manager rights")]
        [TestCase(false, true, true, Description = "NO Administrator, Publication manager and Bpt Manager rights")]
        public void Should_return_true_if_has_bpt_managment_rights(bool isAdministrator, bool isPublicationManager, bool isBptManager)
        {
            MockRights(_publicationId, isAdministrator, isPublicationManager, isBptManager);

            var result = _cmSession.HasBptManagmentRights(_publicationId);
            Assert.IsTrue(result);
        }

        [Test]
        [Description("Should return false if user has no bpt managment rights")]
        [TestCase(false, false, true, Description = "NO Administrator, NO Publication manager and Bpt Manager rights")]
        [TestCase(false, true, false, Description = "NO Administrator, Publication manager and NO Bpt Manager rights")]
        [TestCase(false, false, false, Description = "NO Administrator, NO Publication manager and NO Bpt Manager rights")]
        public void Should_return_false_if_has_no_bpt_managment_rights(bool isAdministrator, bool isPublicationManager, bool isBptManager)
        {
            MockRights(_publicationId, isAdministrator, isPublicationManager, isBptManager);

            var result = _cmSession.HasBptManagmentRights(_publicationId);
            Assert.IsFalse(result);
        }

        [Test]
        [Description("Should return true when it's allowed to edit mappings")]
        [TestCase(true, true, true, Description = "Administrator, Publication manager and Bpt Manager rights")]
        [TestCase(true, false, true, Description = "Administrator, NO Publication manager and Bpt Manager rights")]
        [TestCase(true, true, false, Description = "Administrator, Publication manager and NO Bpt Manager rights")]
        [TestCase(true, false, false, Description = "Administrator, NO Publication manager and NO Bpt Manager rights")]
        [TestCase(false, true, true, Description = "NO Administrator, Publication manager and Bpt Manager rights")]
        [TestCase(false, true, false, Description = "NO Administrator, Publication manager and NO Bpt Manager rights")]
        public void Should_return_true_if_mappings_are_editable(bool isAdministrator, bool isPublicationManager, bool isBptManager)
        {
            MockRights(_publicationId, isAdministrator, isPublicationManager, isBptManager);

            var result = _cmSession.AreMappingsEditable(_publicationId);
            Assert.IsTrue(result);
        }

        [Test]
        [Description("Should return false when it's not allowed to edit mappings")]
        [TestCase(false, false, true, Description = "NO Administrator, NO Publication manager and Bpt Manager rights")]
        [TestCase(false, false, false, Description = "NO Administrator, NO Publication manager and NO Bpt Manager rights")]
        public void Should_return_false_if_mappings_are_not_editables(bool isAdministrator, bool isPublicationManager, bool isBptManager)
        {
            MockRights(_publicationId, isAdministrator, isPublicationManager, isBptManager);

            var result = _cmSession.AreMappingsEditable(_publicationId);
            Assert.IsFalse(result);
        }

        private void MockRights(string publicationId, bool isAdministrator, bool isPublicationManager, bool isBptManager)
        {
            _cmSession.IsAdministrator.Returns(isAdministrator);

            Rights rights = 0;

            if (isPublicationManager)
            {
                rights |= Rights.PublicationManagement;
            }

            if (isBptManager)
            {
                rights |= Rights.BusinessProcessTypeManagement;
            }

            var securityDescriptors = new SecurityDescriptorDataDictionary
            {
                { publicationId, new SecurityDescriptorData{ Rights = rights } }
            };

            _coreServiceClient
                .GetSecurityDescriptorsForSubjectsIds(Arg.Is<string[]>(x => x.SequenceEqual(new[] { publicationId })))
                .Returns(securityDescriptors);
        }
    }
}
