﻿using System;
using System.Collections.Generic;
using NSubstitute;
using NSubstitute.ExceptionExtensions;
using NUnit.Framework;
using SDL.Web.UI.PublicationMappings.Model.Services.Interfaces;
using Tridion.ContentManager.CoreService.Client;
using Tridion.TopologyManager.Client;

namespace SDl.Web.UI.PublicationMappings.Tests.Services.TopologyMappings
{
	[TestFixture]
	public class GetMappingsFixture : TopologyMappingsFixtureBase
	{
		private Func<IEnumerable<MappingData>, string, IEnumerable<MappingData>> _currentMappingsExpander;

		[TestFixtureSetUp]
		public new void TestFixtureSetUp()
		{
			base.TestFixtureSetUp();
			_currentMappingsExpander = TopologyMappingQueryExtensions.MappingsExpander;
		}

		[TestFixtureTearDown]
		public new void TestFixtureTearDown()
		{
			base.TestFixtureTearDown();
			TopologyMappingQueryExtensions.MappingsExpander = _currentMappingsExpander;
		}

		[SetUp]
		public new void SetUp()
		{
			base.SetUp();

			MockRights();
		}

		[Test]
		[Description("Should return topology mappings xml string if user has rights")]
		public void Should_return_topology_mappings_string()
		{
			var mappings = new List<MappingData>()
			{
				MappingData("TestId", "Test web app", "QA", "http://qa.test", "/", PublicationId, "CdEnvironmentId"),
				MappingData("TestId2", "Live web app", "Live", "http://live.net/test", "/test", PublicationId, "CdEnvironmentId"),
				MappingData("TestId3", "Staging web app", "Staging", "http://staging.test", "/", PublicationId, "SomeCdEnvironmentId"),
			};

			MockCdEnvironmentExpand(mappings);

			var actual = TopologyMappingsService.GetMappings(PublicationId);

			const string expected =
				"<pm:Mappings xmlns:pm=\"http://www.sdltridion.com/2015/PublicationMappings\" areMappingsAccessible=\"true\">\r\n" +
				"  <pm:Mapping Id=\"TestId\" WebApplicationId=\"Test web app\" EnvironmentPurpose=\"QA\" BaseMappedUrl=\"http://qa.test\" RelativeUrl=\"/\" PublicationId=\"tcm:0-6-1\" HasPublishedItems=\"false\" CdEnvironmentId=\"CdEnvironmentId\" />\r\n" +
				"  <pm:Mapping Id=\"TestId2\" WebApplicationId=\"Live web app\" EnvironmentPurpose=\"Live\" BaseMappedUrl=\"http://live.net\" RelativeUrl=\"/test\" PublicationId=\"tcm:0-6-1\" HasPublishedItems=\"false\" CdEnvironmentId=\"CdEnvironmentId\" />\r\n" +
				"  <pm:Mapping Id=\"TestId3\" WebApplicationId=\"Staging web app\" EnvironmentPurpose=\"Staging\" BaseMappedUrl=\"http://staging.test\" RelativeUrl=\"/\" PublicationId=\"tcm:0-6-1\" HasPublishedItems=\"false\" CdEnvironmentId=\"SomeCdEnvironmentId\" />\r\n" +
				"</pm:Mappings>";

			Assert.AreEqual(expected, actual);
		}

		[Test]
		[Description("Should return topology mappings xml string only for specific publication")]
		public void Should_return_topology_mappings_only_for_specific_publication()
		{
			var mappings = new List<MappingData>()
			{
				MappingData("TestId", "Test web app", "QA", "http://qa.test", "/", "tcm:0-3-1", "CdEnvironmentId"),
				MappingData("TestId2", "Live web app", "Live", "http://live.net", "/", PublicationId, "CdEnvironmentId"),
				MappingData("TestId3", "Staging web app", "Staging", "http://staging.test", "/", PublicationId, "SomeCdEnvironmentId")
			};

			MockCdEnvironmentExpand(mappings);

			var actual = TopologyMappingsService.GetMappings(PublicationId);

			const string expected =
				"<pm:Mappings xmlns:pm=\"http://www.sdltridion.com/2015/PublicationMappings\" areMappingsAccessible=\"true\">\r\n" +
				"  <pm:Mapping Id=\"TestId2\" WebApplicationId=\"Live web app\" EnvironmentPurpose=\"Live\" BaseMappedUrl=\"http://live.net\" RelativeUrl=\"/\" PublicationId=\"tcm:0-6-1\" HasPublishedItems=\"false\" CdEnvironmentId=\"CdEnvironmentId\" />\r\n" +
				"  <pm:Mapping Id=\"TestId3\" WebApplicationId=\"Staging web app\" EnvironmentPurpose=\"Staging\" BaseMappedUrl=\"http://staging.test\" RelativeUrl=\"/\" PublicationId=\"tcm:0-6-1\" HasPublishedItems=\"false\" CdEnvironmentId=\"SomeCdEnvironmentId\" />\r\n" +
				"</pm:Mappings>";

			Assert.AreEqual(expected, actual);
		}

		[Test]
		[Ignore ("Test has no sense while CRQ-1553 is not fixed")]
		[Description("Should set published state for all mappings")]
		public void Should_set_published_state_for_all_mappings()
		{
			var mappings = new[]
			{
				MappingData("TestId", "Test web app", "QA", "http://qa.test", "/", PublicationId, "CdEnvironmentId"),
				MappingData("TestId2", "Live web app", "Live", "http://live.net/test", "/test", PublicationId, "CdEnvironmentId"),
				MappingData("TestId3", "Staging web app", "Staging", "http://staging.test", "/", PublicationId, "SomeCdEnvironmentId"),
			};

			MockCdEnvironmentExpand(mappings);

			MockPublishedState(mappings[0], true);
			MockPublishedState(mappings[1], false);
			MockPublishedState(mappings[2], true);

			var actual = TopologyMappingsService.GetMappings(PublicationId);

			const string expected =
				"<pm:Mappings xmlns:pm=\"http://www.sdltridion.com/2015/PublicationMappings\" areMappingsAccessible=\"true\">\r\n" +
				"  <pm:Mapping Id=\"TestId\" WebApplicationId=\"Test web app\" EnvironmentPurpose=\"QA\" BaseMappedUrl=\"http://qa.test\" RelativeUrl=\"/\" PublicationId=\"tcm:0-6-1\" HasPublishedItems=\"true\" CdEnvironmentId=\"CdEnvironmentId\" />\r\n" +
				"  <pm:Mapping Id=\"TestId2\" WebApplicationId=\"Live web app\" EnvironmentPurpose=\"Live\" BaseMappedUrl=\"http://live.net\" RelativeUrl=\"/test\" PublicationId=\"tcm:0-6-1\" HasPublishedItems=\"false\" CdEnvironmentId=\"CdEnvironmentId\" />\r\n" +
				"  <pm:Mapping Id=\"TestId3\" WebApplicationId=\"Staging web app\" EnvironmentPurpose=\"Staging\" BaseMappedUrl=\"http://staging.test\" RelativeUrl=\"/\" PublicationId=\"tcm:0-6-1\" HasPublishedItems=\"true\" CdEnvironmentId=\"SomeCdEnvironmentId\" />\r\n" +
				"</pm:Mappings>";

			Assert.AreEqual(expected, actual);
		}

		[Test]
		[Description("Should set areMappingsAccessible to false if an error occured during getting mappings")]
		public void Should_throw_exception_when_get_mappings_string()
		{
			TopologyManagerClient.Mappings.Throws(new InvalidOperationException());

			CmSession.IsAdministrator.Returns(true);

			var actual = TopologyMappingsService.GetMappings(PublicationId);

			const string expected =
				"<pm:Mappings xmlns:pm=\"http://www.sdltridion.com/2015/PublicationMappings\" areMappingsAccessible=\"false\" />";

			Assert.AreEqual(expected, actual);
		}

		private static MappingData MappingData(string id, string webApplicationId, string environmentPurpose, string primaryMappedUrl, string relativeUrl, string publicationId, string cdEnvironmentId)
		{
			return new MappingData
			{
				Id = id,
				WebApplicationId = webApplicationId,
				EnvironmentPurpose = environmentPurpose,
				PrimaryMappedUrl = primaryMappedUrl,
				RelativeUrl = relativeUrl,
				PublicationId = publicationId,
				CdEnvironment = new CdEnvironmentData { Id = cdEnvironmentId },
				IsOffline = null,
				CmEnvironmentId = Tridion.TopologyManager.Client.TopologyManagerClient.ContentManagerEnvironmentId
			};
		}

		private static void MockCdEnvironmentExpand(IEnumerable<MappingData> mappings)
		{
			TopologyMappingQueryExtensions.MappingsExpander = (data, path) => mappings;
		}

		private void MockPublishedState(MappingData mapping, bool hasPublishedItems = false)
		{
			var fakeList = hasPublishedItems
				? new[] {new IdentifiableObjectData()}
				: new IdentifiableObjectData[0];

			// TODO AN: temporaly check if publication has any published items due to CRQ-1553
//			CoreService
//				.GetSystemWideList(Arg.Is<PublishedItemsFilterData>(
//					x => x.Purpose == mapping.EnvironmentPurpose && x.ForPublication.IdRef == mapping.PublicationId))
//				.Returns(fakeList);
			CoreService
				.GetSystemWideList(Arg.Is<PublishedItemsFilterData>(
					x => x.ForPublication.IdRef == mapping.PublicationId))
				.Returns(fakeList);
		}
	}
}
