﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using NSubstitute;
using NSubstitute.ExceptionExtensions;
using NUnit.Framework;
using SDL.Web.UI.PublicationMappings.Model.Services.Interfaces;
using Tridion.TopologyManager.Client;

namespace SDl.Web.UI.PublicationMappings.Tests.Services.TopologyMappings
{
	[TestFixture]
	public class GetTopologiesFixture : TopologyMappingsFixtureBase
	{
		private const string CdTopologyTypeId = "CDTopologyType";
		private const string AnotherTopologyTypeId = "AnotherTopologyType";

		private Func<IEnumerable<WebApplicationData>, string, IEnumerable<WebApplicationData>> _currentWebApplicationExpander;

		private readonly IEnumerable<WebApplicationData> _webApps = new[]
		{
			WebApplication("WebApp1", "CdEnv1"),
			WebApplication("WebApp2", "CdEnv1"),
			WebApplication("WebApp3", "CdEnv2"),
			WebApplication("WebApp4", "CdEnv2"),
			WebApplication("WebApp5", "CdEnv3"),
			WebApplication("WebApp6", "CdEnv3"),
		};

		private readonly IEnumerable<CdTopologyData> _topologies = new[]
		{
			Topology("CdTopologyId1", "First Topology", CdTopologyTypeId, new Collection<string> {"CdEnv1", "CdEnv2"}),
			Topology("CdTopologyIdAlternative", "Alternative Topology", AnotherTopologyTypeId, new Collection<string> {"CdEnv1", "CdEnv2"}),
			Topology("CdTopologyId2", "Second Topology", CdTopologyTypeId, new Collection<string> {"CdEnv2", "CdEnv3"}),
			Topology("CdTopologyId3", "Third Topology", CdTopologyTypeId, new Collection<string> {"CdEnv4", "CdEnv5"})
		};

		[TestFixtureSetUp]
		public new void TestFixtureSetUp()
		{
			base.TestFixtureSetUp();
			_currentWebApplicationExpander = TopologyMappingQueryExtensions.WebApplicationExpander;
		}

		[TestFixtureTearDown]
		public new void TestFixtureTearDown()
		{
			base.TestFixtureTearDown();
			TopologyMappingQueryExtensions.WebApplicationExpander = _currentWebApplicationExpander;
		}

		[SetUp]
		public new void SetUp()
		{
			base.SetUp();

			MockCdTopologies(_topologies);
			MockWebApplications(_webApps);
		}

		[Test]
		public void Should_return_topologies()
		{
			var webApplicationIds = new[]
			{
				_webApps.ElementAt(0).Id,
				_webApps.ElementAt(3).Id
			};

			var actual = TopologyMappingsService.GetTopologies(CdTopologyTypeId, webApplicationIds);

			const string expected =
				"<pm:CdTopologies xmlns:pm=\"http://www.sdltridion.com/2015/PublicationMappings\">\r\n" +
				"  <pm:CdTopology Id=\"CdTopologyId1\" Name=\"First Topology\">\r\n" +
				"    <pm:CdEnvironment Id=\"CdEnv1\" />\r\n" +
				"    <pm:CdEnvironment Id=\"CdEnv2\" />\r\n" +
				"  </pm:CdTopology>\r\n" +
				"  <pm:CdTopology Id=\"CdTopologyId2\" Name=\"Second Topology\">\r\n" +
				"    <pm:CdEnvironment Id=\"CdEnv2\" />\r\n" +
				"    <pm:CdEnvironment Id=\"CdEnv3\" />\r\n" +
				"  </pm:CdTopology>\r\n" +
				"</pm:CdTopologies>";

			Assert.AreEqual(expected, actual);
		}

		[Test]
		public void Should_return_all_topologies_when_webApps_are_not_specified()
		{
			var actual = TopologyMappingsService.GetTopologies(CdTopologyTypeId);

			const string expected =
				"<pm:CdTopologies xmlns:pm=\"http://www.sdltridion.com/2015/PublicationMappings\">\r\n" +
				"  <pm:CdTopology Id=\"CdTopologyId1\" Name=\"First Topology\">\r\n" +
				"    <pm:CdEnvironment Id=\"CdEnv1\" />\r\n" +
				"    <pm:CdEnvironment Id=\"CdEnv2\" />\r\n" +
				"  </pm:CdTopology>\r\n" +
				"  <pm:CdTopology Id=\"CdTopologyId2\" Name=\"Second Topology\">\r\n" +
				"    <pm:CdEnvironment Id=\"CdEnv2\" />\r\n" +
				"    <pm:CdEnvironment Id=\"CdEnv3\" />\r\n" +
				"  </pm:CdTopology>\r\n" +
				"  <pm:CdTopology Id=\"CdTopologyId3\" Name=\"Third Topology\">\r\n" +
				"    <pm:CdEnvironment Id=\"CdEnv4\" />\r\n" +
				"    <pm:CdEnvironment Id=\"CdEnv5\" />\r\n" +
				"  </pm:CdTopology>\r\n" +
				"</pm:CdTopologies>";

			Assert.AreEqual(expected, actual);
		}

		[Test]
		public void Should_return_topologies_according_to_topology_type_id()
		{
			var actual = TopologyMappingsService.GetTopologies(AnotherTopologyTypeId);

			const string expected =
				"<pm:CdTopologies xmlns:pm=\"http://www.sdltridion.com/2015/PublicationMappings\">\r\n" +
				"  <pm:CdTopology Id=\"CdTopologyIdAlternative\" Name=\"Alternative Topology\">\r\n" +
				"    <pm:CdEnvironment Id=\"CdEnv1\" />\r\n" +
				"    <pm:CdEnvironment Id=\"CdEnv2\" />\r\n" +
				"  </pm:CdTopology>\r\n" +
				"</pm:CdTopologies>";

			Assert.AreEqual(expected, actual);
		}

		[Test]
		public void Should_handle_Topology_Manager_exception()
		{
			const string errorMessage = "Can not get topologies";

			TopologyManagerClient.CdTopologies.Throws(new InvalidOperationException(errorMessage));

			var exception = Assert.Throws<TopologyManagerClient.DataServiceException>(() => TopologyMappingsService.GetTopologies(CdTopologyTypeId));
			Assert.AreEqual(errorMessage, exception.Message);
		}

		private static WebApplicationData WebApplication(string id, string cdEnvironmentId)
		{
			return new WebApplicationData
			{
				Id = id,
				CdEnvironment = new CdEnvironmentData { Id = cdEnvironmentId }
			};
		}

		private static CdTopologyData Topology(string topologyid, string topologyName, string cdTopologyTypeId, Collection<string> cdEnvironmentIds)
		{
			return new CdTopologyData
			{
				Id = topologyid,
				CdEnvironmentIds = cdEnvironmentIds,
				Name = topologyName,
				CdTopologyTypeId = cdTopologyTypeId
			};
		}

		private void MockCdTopologies(IEnumerable<CdTopologyData> topologies)
		{
			TopologyManagerClient.CdTopologies.Returns(topologies);
		}

		private static void MockWebApplications(IEnumerable<WebApplicationData> webApps)
		{
			TopologyMappingQueryExtensions.WebApplicationExpander = (data, path) => webApps;
		}
	}
}
