﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using NSubstitute;
using NSubstitute.ExceptionExtensions;
using NUnit.Framework;
using SDL.Web.UI.PublicationMappings.Model.Services.Interfaces;
using Tridion.TopologyManager.Client;

namespace SDl.Web.UI.PublicationMappings.Tests.Services.TopologyMappings
{
	[TestFixture]
	public class GetWebAppsFixture : TopologyMappingsFixtureBase
	{
		private Func<IEnumerable<WebApplicationData>, string, IEnumerable<WebApplicationData>> _currentWebApplicationExpander;

		private readonly IEnumerable<WebApplicationData> _webApps = new[]
		{
			WebApplication("WebApp1", "WebSite1", "/", "CdEnv1", "Staging"),
			WebApplication("WebApp2", "WebSite1", "/test", "CdEnv1", "Staging"),
			WebApplication("WebApp3", "WebSite1", "/new", "CdEnv1", "Staging"),
			WebApplication("WebApp4", "WebSite2", "/123", "CdEnv2", "Live"),
			WebApplication("WebApp5", "WebSite2", "/abc", "CdEnv2", "Live"),
			WebApplication("WebApp6", "WebSite3", "/third", "CdEnv1", "Staging"),
			WebApplication("WebApp7", "WebSiteAlternative", "/alternative", "CdEnv3", "Live"),
		};

		private readonly IEnumerable<WebsiteData> _webSites = new[]
		{
			WebSite("WebSite1", "http://first.web.site"),
			WebSite("WebSite2", "http://second.web.site"),
			WebSite("WebSite3", "http://another.web.site"),
			WebSite("WebSiteAlternative", "http://alternative.web.site"),
		};

		[TestFixtureSetUp]
		public new void TestFixtureSetUp()
		{
			base.TestFixtureSetUp();
			_currentWebApplicationExpander = TopologyMappingQueryExtensions.WebApplicationExpander;
		}

		[TestFixtureTearDown]
		public new void TestFixtureTearDown()
		{
			base.TestFixtureTearDown();
			TopologyMappingQueryExtensions.WebApplicationExpander = _currentWebApplicationExpander;
		}

		[SetUp]
		public new void SetUp()
		{
			base.SetUp();

			MockWebApplications(_webApps);
			MockWebSites(_webSites);
		}

		[Test]
		public void Should_return_all_web_applications()
		{
			var environmentIds = new[] { "CdEnv1", "CdEnv2", "CdEnv3" };

			var actual = TopologyMappingsService.GetWebApplications(environmentIds);

			const string expected =
				"<pm:WebApplications xmlns:pm=\"http://www.sdltridion.com/2015/PublicationMappings\">\r\n" +
				"  <pm:WebApplication Id=\"WebApp1\">http://first.web.site/</pm:WebApplication>\r\n" +
				"  <pm:WebApplication Id=\"WebApp2\">http://first.web.site/test</pm:WebApplication>\r\n" +
				"  <pm:WebApplication Id=\"WebApp3\">http://first.web.site/new</pm:WebApplication>\r\n" +
				"  <pm:WebApplication Id=\"WebApp4\">http://second.web.site/123</pm:WebApplication>\r\n" +
				"  <pm:WebApplication Id=\"WebApp5\">http://second.web.site/abc</pm:WebApplication>\r\n" +
				"  <pm:WebApplication Id=\"WebApp6\">http://another.web.site/third</pm:WebApplication>\r\n" +
				"  <pm:WebApplication Id=\"WebApp7\">http://alternative.web.site/alternative</pm:WebApplication>\r\n" +
				"</pm:WebApplications>";

			Assert.AreEqual(expected, actual);
		}

		[Test]
		public void Should_return_filtered_web_applications_by_environment_ids()
		{
			var environmentIds = new[] { "CdEnv1" };

			var actual = TopologyMappingsService.GetWebApplications(environmentIds);

			const string expected =
				"<pm:WebApplications xmlns:pm=\"http://www.sdltridion.com/2015/PublicationMappings\">\r\n" +
				"  <pm:WebApplication Id=\"WebApp1\">http://first.web.site/</pm:WebApplication>\r\n" +
				"  <pm:WebApplication Id=\"WebApp2\">http://first.web.site/test</pm:WebApplication>\r\n" +
				"  <pm:WebApplication Id=\"WebApp3\">http://first.web.site/new</pm:WebApplication>\r\n" +
				"  <pm:WebApplication Id=\"WebApp6\">http://another.web.site/third</pm:WebApplication>\r\n" +
				"</pm:WebApplications>";
			Console.Write(actual);
			Assert.AreEqual(expected, actual);
		}


		[Test]
		public void Should_return_filtered_web_applications_by_environment_ids_and_purpose()
		{
			var environmentIds = new[] { "CdEnv1", "CdEnv3" };

			var actual = TopologyMappingsService.GetWebApplications(environmentIds, "Live");

			const string expected =
				"<pm:WebApplications xmlns:pm=\"http://www.sdltridion.com/2015/PublicationMappings\">\r\n" +
				"  <pm:WebApplication Id=\"WebApp7\">http://alternative.web.site/alternative</pm:WebApplication>\r\n" +
				"</pm:WebApplications>";
			Console.Write(actual);
			Assert.AreEqual(expected, actual);
		}

		[Test]
		public void Should_handle_Topology_Manager_exception()
		{
			const string errorMessage = "Can not get web applications";

			TopologyManagerClient.WebApplications.Throws(new InvalidOperationException(errorMessage));

			var exception = Assert.Throws<TopologyManagerClient.DataServiceException>(() => TopologyMappingsService.GetWebApplications(new[] { "" }));
			Assert.AreEqual(errorMessage, exception.Message);
		}

		private static WebApplicationData WebApplication(string id, string webSiteId, string contextUrl, string cdEnvironmentId, string cdEnvironmentPurpose)
		{
			return new WebApplicationData
			{
				Id = id,
				WebsiteId = webSiteId,
				ContextUrl = contextUrl,
				EnvironmentPurpose = cdEnvironmentPurpose,
				CdEnvironment = new CdEnvironmentData { Id = cdEnvironmentId, EnvironmentPurpose = cdEnvironmentPurpose }
			};
		}

		private static WebsiteData WebSite(string id, string baseUrl)
		{
			return new WebsiteData
			{
				Id = id,
				BaseUrls = new Collection<string> { baseUrl, "http:\\some.other.url" }
			};
		}

		private void MockWebApplications(IEnumerable<WebApplicationData> webApps)
		{
			TopologyMappingQueryExtensions.WebApplicationExpander = (data, path) => webApps;
		}

		private void MockWebSites(IEnumerable<WebsiteData> webSites)
		{
			TopologyManagerClient.Websites.Returns(webSites);
		}
	}
}
